#!/usr/bin/python
import setuptools

setuptools.setup(
    name='doc2json',
    version='0.1',
    packages=setuptools.find_packages(),
    install_requires=[
    ],
    tests_require=[
    ],
    zip_safe=False,
    test_suite='py.test',
    entry_points='',
)